#include <windows.h>
#include <gl/gl.h>
#include "glide.h"
#include "main.h"
#include "resource.h"

typedef struct _wrapper_config
{
	int res;
	int filter;
	int disable_glsl;
	int disable_dithered_alpha;
} wrapper_config;

static HINSTANCE hinstance;
static HKEY Key = NULL;
static wrapper_config config;
static int valid_filter = 0;
static int valid_dithalpha = 0;

BOOL WINAPI DllMain(
  HINSTANCE hinstDLL,
  DWORD fdwReason,
  LPVOID lpvReserved
)
{
	hinstance = hinstDLL;
	return TRUE;
}

static BOOL registry_open()
{
	DWORD Status;
	LONG Res;
	
	Res = RegCreateKeyEx(HKEY_LOCAL_MACHINE, "SOFTWARE\\GlideWrapper", 0, NULL, 
		                 REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &Key, &Status);
	if(Res != ERROR_SUCCESS) return FALSE;
	return TRUE;
}

static BOOL registry_getvalue(char *name, char *buffer, unsigned long size)
{
	LONG Res = RegQueryValueEx(Key, name, NULL, NULL, (unsigned char *)buffer, &size);
	if(Res != ERROR_SUCCESS) return FALSE;
	return TRUE;
}

static BOOL registry_setvalue(char *name, DWORD type, char *buffer, unsigned long size)
{
	LONG Res = RegSetValueEx(Key, name, 0, type, (unsigned char *)buffer, size);
	if(Res != ERROR_SUCCESS) return FALSE;
	return TRUE;
}

static void registry_close()
{
	RegCloseKey(Key);
	Key = NULL;
}

static void readConfig()
{
	registry_open();
	if (registry_getvalue("res", (char*)&config.res, sizeof(int)) == FALSE)
		config.res = 7;
	if (registry_getvalue("filter", (char*)&config.filter, sizeof(int)) == FALSE)
		config.filter = 0;
	if (registry_getvalue("glsl", (char*)&config.disable_glsl, sizeof(int)) == FALSE)
		config.disable_glsl = 0;
	if (registry_getvalue("dithalpha", (char*)&config.disable_dithered_alpha, sizeof(int)) == FALSE)
		config.disable_dithered_alpha = 0;
	registry_close();
}

static void saveConfig()
{
	registry_open();
	registry_setvalue("res", REG_DWORD, (char*)&config.res, sizeof(int));
	registry_setvalue("filter", REG_DWORD, (char*)&config.filter, sizeof(int));
	registry_setvalue("glsl", REG_DWORD, (char*)&config.disable_glsl, sizeof(int));
	registry_setvalue("dithalpha", REG_DWORD, (char*)&config.disable_dithered_alpha, sizeof(int));
	registry_close();
}

static BOOL CALLBACK ConfigDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{
	switch(Message)
    {
	case WM_INITDIALOG:
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"320x200");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"320x240");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"400x256");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"512x384");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"640x200");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"640x350");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"640x400");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"640x480");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"800x600");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"960x720");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"856x480");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"512x256");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"1024x768");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"1280x1024");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"1600x1200");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"400x300"); 
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"1152x864"); 
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"1280x960");                          
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"1600x1024");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"1792x1344");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"1856x1392");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"1920x1440");
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_ADDSTRING, 0, (LPARAM)"2048x1536");
		

		SendDlgItemMessage(hwnd, IDC_FILTER_COMBO, CB_ADDSTRING, 0, (LPARAM)"None");
		SendDlgItemMessage(hwnd, IDC_FILTER_COMBO, CB_ADDSTRING, 0, (LPARAM)"Blur edges");
		SendDlgItemMessage(hwnd, IDC_FILTER_COMBO, CB_ADDSTRING, 0, (LPARAM)"Super 2xSai");
		SendDlgItemMessage(hwnd, IDC_FILTER_COMBO, CB_ADDSTRING, 0, (LPARAM)"HQ2X");
		SendDlgItemMessage(hwnd, IDC_FILTER_COMBO, CB_ADDSTRING, 0, (LPARAM)"HQ4X");

		readConfig();
		SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_SETCURSEL, config.res, 0);
		SendDlgItemMessage(hwnd, IDC_FILTER_COMBO, CB_SETCURSEL, config.filter, 0);
		if(config.disable_glsl)
			SendDlgItemMessage(hwnd, IDC_GLSL, BM_SETCHECK, BST_CHECKED, 0);
		else
			SendDlgItemMessage(hwnd, IDC_GLSL, BM_SETCHECK, BST_UNCHECKED, 0);
		if(config.disable_dithered_alpha)
			SendDlgItemMessage(hwnd, IDC_DITHALPHA, BM_SETCHECK, BST_CHECKED, 0);
		else
			SendDlgItemMessage(hwnd, IDC_DITHALPHA, BM_SETCHECK, BST_UNCHECKED, 0);
		return TRUE;
		break;

	case WM_CLOSE:
		EndDialog(hwnd, IDOK);
		break;

	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case IDCANCEL:
			EndDialog(hwnd, IDOK);
			break;
		case IDOK:
			config.res = SendDlgItemMessage(hwnd, IDC_RES_COMBO, CB_GETCURSEL, 0, 0);
			config.filter = SendDlgItemMessage(hwnd, IDC_FILTER_COMBO, CB_GETCURSEL, 0, 0);
			config.disable_glsl = SendDlgItemMessage(hwnd, IDC_GLSL, BM_GETCHECK,0,0) == BST_CHECKED ? 1 : 0;
			config.disable_dithered_alpha = SendDlgItemMessage(hwnd, IDC_DITHALPHA, BM_GETCHECK,0,0) == BST_CHECKED ? 1 : 0;
			valid_filter = 0;
			saveConfig();
			EndDialog(hwnd, IDOK);
			break;
		}
		break;

	default:
		return FALSE;
    }
    return TRUE;
}

FX_ENTRY void FX_CALL grConfigWrapperExt(HINSTANCE
instance, HWND hwnd)
{
	DialogBox(hinstance, MAKEINTRESOURCE(IDD_CONFIG_DIAL), hwnd, ConfigDlgProc);
}

int getFullScreenWidth()
{
	readConfig();
	switch(config.res)
	{
	case GR_RESOLUTION_320x200:
		return 320;
		break;
	case GR_RESOLUTION_320x240:
		return 320;
		break;
	case GR_RESOLUTION_400x256:
		return 400;
		break;
	case GR_RESOLUTION_512x384:
		return 512;
		break;
	case GR_RESOLUTION_640x200:
		return 640;
		break;
	case GR_RESOLUTION_640x350:
		return 640;
		break;
	case GR_RESOLUTION_640x400:
		return 640;
		break;
	case GR_RESOLUTION_640x480:
		return 640;
		break;
	case GR_RESOLUTION_800x600:
		return 800;
		break;
	case GR_RESOLUTION_960x720:
		return 960;
		break;
	case GR_RESOLUTION_856x480:
		return 856;
		break;
	case GR_RESOLUTION_512x256:
		return 512;
		break;
	case GR_RESOLUTION_1024x768:
		return 1024;
		break;
	case GR_RESOLUTION_1280x1024:
		return 1280;
		break;
	case GR_RESOLUTION_1600x1200:
		return 1600;
		break;
	case GR_RESOLUTION_400x300:
		return 400;
		break;
	case GR_RESOLUTION_1152x864:
		return 1152;
		break;
	case GR_RESOLUTION_1280x960:
		return 1280;
		break;
	case GR_RESOLUTION_1600x1024:
		return 1600;
		break;
	case GR_RESOLUTION_1792x1344:
		return 1792;
		break;
    case GR_RESOLUTION_1856x1392:
		return 1856;
		break;
    case GR_RESOLUTION_1920x1440:
		return 1920;
		break;
    case GR_RESOLUTION_2048x1536:
		return 2048;
		break;
	
	}
	return 0;
}

int getFullScreenHeight()
{
	readConfig();
	switch(config.res)
	{
	case GR_RESOLUTION_320x200:
		return 200;
		break;
	case GR_RESOLUTION_320x240:
		return 240;
		break;
	case GR_RESOLUTION_400x256:
		return 256;
		break;
	case GR_RESOLUTION_512x384:
		return 384;
		break;
	case GR_RESOLUTION_640x200:
		return 200;
		break;
	case GR_RESOLUTION_640x350:
		return 350;
		break;
	case GR_RESOLUTION_640x400:
		return 400;
		break;
	case GR_RESOLUTION_640x480:
		return 480;
		break;
	case GR_RESOLUTION_800x600:
		return 600;
		break;
	case GR_RESOLUTION_960x720:
		return 720;
		break;
	case GR_RESOLUTION_856x480:
		return 480;
		break;
	case GR_RESOLUTION_512x256:
		return 256;
		break;
	case GR_RESOLUTION_1024x768:
		return 768;
		break;
	case GR_RESOLUTION_1280x1024:
		return 1024;
		break;
	case GR_RESOLUTION_1600x1200:
		return 1200;
		break;
	case GR_RESOLUTION_400x300:
		return 300;
		break;
	case GR_RESOLUTION_1152x864:
		return 864;
		break;
	case GR_RESOLUTION_1280x960:
		return 960;
		break;
	case GR_RESOLUTION_1600x1024:
		return 1024;
		break;
	case GR_RESOLUTION_1792x1344:
		return 1344;
		break;
   case GR_RESOLUTION_1856x1392:
		return 1392;
		break;
   case GR_RESOLUTION_1920x1440:
		return 1440;
		break;
   case GR_RESOLUTION_2048x1536:
		return 1536;
		break;
	}
	return 0;
}

FX_ENTRY GrScreenResolution_t FX_CALL grWrapperFullScreenResolutionExt(void)
{
	readConfig();
	return config.res;
}

int getFilter()
{
	if (!valid_filter)
	{
		readConfig();
		valid_filter = 1;
	}
	return config.filter;
}

int getDisableDitheredAlpha()
{
	if (!valid_dithalpha)
	{
		readConfig();
		valid_dithalpha = 1;
	}
	return config.disable_dithered_alpha;
}

int getDisableGLSL()
{
	readConfig();
	return config.disable_glsl;
}
