#include <windows.h>
#include <gl/gl.h>
#include "glide.h"
#include "main.h"

/* Napalm extensions to GrTextureFormat_t */
#define GR_TEXFMT_ARGB_CMP_FXT1           0x11
#define GR_TEXFMT_ARGB_8888               0x12
#define GR_TEXFMT_YUYV_422                0x13
#define GR_TEXFMT_UYVY_422                0x14
#define GR_TEXFMT_AYUV_444                0x15
#define GR_TEXFMT_ARGB_CMP_DXT1           0x16
#define GR_TEXFMT_ARGB_CMP_DXT2           0x17
#define GR_TEXFMT_ARGB_CMP_DXT3           0x18
#define GR_TEXFMT_ARGB_CMP_DXT4           0x19
#define GR_TEXFMT_ARGB_CMP_DXT5           0x1A
#define GR_TEXTFMT_RGB_888                0xFF

#define TMU_SIZE 8*1024*1024

int tex0_width, tex0_height, tex1_width, tex1_height;
float lambda;

static int min_filter0, mag_filter0, wrap_s0, wrap_t0;
static int min_filter1, mag_filter1, wrap_s1, wrap_t1;

unsigned char *filter(unsigned char *source, int width, int height, int *width2, int *height2);

void init_textures()
{
	tex0_width = tex0_height = tex1_width = tex1_height = 2;	
}


FX_ENTRY FxU32 FX_CALL 
grTexMinAddress( GrChipID_t tmu )
{
	LOG("grTexMinAddress(%d)\r\n", tmu);
	return tmu*TMU_SIZE;
}

FX_ENTRY FxU32 FX_CALL 
grTexMaxAddress( GrChipID_t tmu )
{
	LOG("grTexMaxAddress(%d)\r\n", tmu);
	return tmu*TMU_SIZE + TMU_SIZE - 1;
}

FX_ENTRY FxU32 FX_CALL 
grTexTextureMemRequired( FxU32     evenOdd,
                                 GrTexInfo *info   )
{
	int width, height;
	LOG("grTextureMemRequired(%d)\r\n", evenOdd);
	if (info->largeLodLog2 != info->smallLodLog2) display_warning("grTexTextureMemRequired : loading more than one LOD");

	if (info->aspectRatioLog2 < 0)
	{
		height = 1 << info->largeLodLog2;
		width = height >> -info->aspectRatioLog2;
	}
	else
	{
		width = 1 << info->largeLodLog2;
		height = width >> info->aspectRatioLog2;
	}

	switch(info->format)
	{
	case GR_TEXFMT_ALPHA_8:
	case GR_TEXFMT_ALPHA_INTENSITY_44:
		return width*height;
		break;
	case GR_TEXFMT_ARGB_1555:
	case GR_TEXFMT_ARGB_4444:
	case GR_TEXFMT_ALPHA_INTENSITY_88:
		return width*height*2;
		break;
	default:
		display_warning("grTexTextureMemRequired : unknown texture format: %x", info->format);
	}
	return 0;
}

FX_ENTRY FxU32 FX_CALL 
grTexCalcMemRequired(
                     GrLOD_t lodmin, GrLOD_t lodmax,
                     GrAspectRatio_t aspect, GrTextureFormat_t fmt)
{
	int width, height;
	LOG("grTexCalcMemRequired(%d, %d, %d, %d)\r\n", lodmin, lodmax, aspect, fmt);
	if (lodmax != lodmin) display_warning("grTexCalcMemRequired : loading more than one LOD");

	if (aspect < 0)
	{
		height = 1 << lodmax;
		width = height >> -aspect;
	}
	else
	{
		width = 1 << lodmax;
		height = width >> aspect;
	}

	switch(fmt)
	{
	case GR_TEXFMT_ALPHA_8:
	case GR_TEXFMT_ALPHA_INTENSITY_44:
		return width*height;
		break;
	case GR_TEXFMT_ARGB_1555:
	case GR_TEXFMT_ARGB_4444:
	case GR_TEXFMT_ALPHA_INTENSITY_88:
	case GR_TEXFMT_RGB_565:
		return width*height*2;
		break;
	default:
		display_warning("grTexTextureMemRequired : unknown texture format: %x", fmt);
	}
	return 0;
}

FX_ENTRY void FX_CALL 
grTexDownloadMipMap( GrChipID_t tmu,
                     FxU32      startAddress,
                     FxU32      evenOdd,
                     GrTexInfo  *info )
{
	int width, height, i, j;
	unsigned char* texture = NULL;
	unsigned char* filtered_texture = NULL;
	int factor = 0;
	LOG("grTexDownloadMipMap(%d,%d,%d)\r\n", tmu, startAddress, evenOdd);
	if (info->largeLodLog2 != info->smallLodLog2) display_warning("grTexDownloadMipMap : loading more than one LOD");

	if (info->aspectRatioLog2 < 0)
	{
		height = 1 << info->largeLodLog2;
		width = height >> -info->aspectRatioLog2;
	}
	else
	{
		width = 1 << info->largeLodLog2;
		height = width >> info->aspectRatioLog2;
	}

	texture = malloc(width*height*4);
	switch(info->format)
	{
	case GR_TEXFMT_ALPHA_8:
		for (j=0; j<width; j++)
		{
			for (i=0; i<height; i++)
			{
				texture[i*width*4+j*4+0]= ((unsigned char*)info->data)[i*width+j];
				texture[i*width*4+j*4+1]= ((unsigned char*)info->data)[i*width+j];
				texture[i*width*4+j*4+2]= ((unsigned char*)info->data)[i*width+j];
				texture[i*width*4+j*4+3]= ((unsigned char*)info->data)[i*width+j];
			}
		}
		factor = 1;
		break;
	case GR_TEXFMT_ALPHA_INTENSITY_44:
		for (j=0; j<width; j++)
		{
			for (i=0; i<height; i++)
			{
				texture[i*width*4+j*4+3]= (((unsigned char*)info->data)[i*width+j]&0xF0);
				texture[i*width*4+j*4+0]= ((((unsigned char*)info->data)[i*width+j]&0xF)<<4);
				texture[i*width*4+j*4+1]= ((((unsigned char*)info->data)[i*width+j]&0xF)<<4);
				texture[i*width*4+j*4+2]= ((((unsigned char*)info->data)[i*width+j]&0xF)<<4);
			}
		}
		factor = 1;
		break;
	case GR_TEXFMT_RGB_565:
		for (j=0; j<width; j++)
		{
			for (i=0; i<height; i++)
			{
				texture[i*width*4+j*4+3] = 0;
				texture[i*width*4+j*4+0]=((((unsigned short*)info->data)[i*width+j]>>11)&0x1F)<<3;
				texture[i*width*4+j*4+1]=((((unsigned short*)info->data)[i*width+j]>> 5)&0x3F)<<2;
				texture[i*width*4+j*4+2]=((((unsigned short*)info->data)[i*width+j]>> 0)&0x1F)<<3;
			}
		}
		factor = 2;
		break;
	case GR_TEXFMT_ARGB_1555:
		for (j=0; j<width; j++)
		{
			for (i=0; i<height; i++)
			{
				texture[i*width*4+j*4+3]=(((unsigned short*)info->data)[i*width+j]>>15)!=0 ? 0xFF : 0;
				texture[i*width*4+j*4+0]=((((unsigned short*)info->data)[i*width+j]>>10)&0x1F)<<3;
				texture[i*width*4+j*4+1]=((((unsigned short*)info->data)[i*width+j]>> 5)&0x1F)<<3;
				texture[i*width*4+j*4+2]=((((unsigned short*)info->data)[i*width+j]>> 0)&0x1F)<<3;
			}
		}
		factor = 2;
		break;
	case GR_TEXFMT_ALPHA_INTENSITY_88:
		for (j=0; j<width; j++)
		{
			for (i=0; i<height; i++)
			{
				texture[i*width*4+j*4+3]= ((unsigned char*)info->data)[i*width*2+j*2+1];
				texture[i*width*4+j*4+0]= ((unsigned char*)info->data)[i*width*2+j*2];
				texture[i*width*4+j*4+1]= ((unsigned char*)info->data)[i*width*2+j*2];
				texture[i*width*4+j*4+2]= ((unsigned char*)info->data)[i*width*2+j*2];
			}
		}
		factor = 2;
		break;
	case GR_TEXFMT_ARGB_4444:
		for (j=0; j<width; j++)
		{
			for (i=0; i<height; i++)
			{
				texture[i*width*4+j*4+3]=((((unsigned short*)info->data)[i*width+j]>>12)&0xF)<<4;
				texture[i*width*4+j*4+0]=((((unsigned short*)info->data)[i*width+j]>> 8)&0xF)<<4;
				texture[i*width*4+j*4+1]=((((unsigned short*)info->data)[i*width+j]>> 4)&0xF)<<4;
				texture[i*width*4+j*4+2]=((((unsigned short*)info->data)[i*width+j]>> 0)&0xF)<<4;
			}
		}
		factor = 2;
		break;
	case GR_TEXFMT_ARGB_8888:
		for (j=0; j<width; j++)
		{
			for (i=0; i<height; i++)
			{
				texture[i*width*4+j*4+3]= ((unsigned char*)info->data)[i*width*4+j*4+3];
				texture[i*width*4+j*4+0]= ((unsigned char*)info->data)[i*width*4+j*4+2];
				texture[i*width*4+j*4+1]= ((unsigned char*)info->data)[i*width*4+j*4+1];
				texture[i*width*4+j*4+2]= ((unsigned char*)info->data)[i*width*4+j*4+0];
			}
		}
		factor = 4;
		break;
	default:
		display_warning("grTexDownloadMipMap : unknown texture format: %x", info->format);
	}
	if (nbTextureUnits <= 2)
		glActiveTextureARB(GL_TEXTURE1_ARB);
	else
		glActiveTextureARB(GL_TEXTURE2_ARB);
	glDeleteTextures(1 ,&startAddress);
	glGenTextures(1, &startAddress+1);
	glBindTexture(GL_TEXTURE_2D, startAddress+1);
	if (getFilter() == 0)
		glTexImage2D(GL_TEXTURE_2D, 0, 4, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, texture);
	else
	{
		int width2, height2;
		filtered_texture = filter(texture, width, height, &width2, &height2);
		glTexImage2D(GL_TEXTURE_2D, 0, 4, width2, height2, 0, GL_RGBA, GL_UNSIGNED_BYTE, filtered_texture);
	}
	glBindTexture(GL_TEXTURE_2D, 32*1024*1024); //fixes hi-res on puzzle FB effect in BT in windowed mode
	if (texture) free(texture);
	if (filtered_texture) free(filtered_texture);
}

CheckTextureBufferFormat(GrChipID_t tmu, FxU32 startAddress, GrTexInfo *info );
FX_ENTRY void FX_CALL 
grTexSource( GrChipID_t tmu,
             FxU32      startAddress,
             FxU32      evenOdd,
             GrTexInfo  *info )
{
	LOG("grTexSource(%d,%d,%d)\r\n", tmu, startAddress, evenOdd);

	
	if (tmu == GR_TMU1 || nbTextureUnits <= 2)
	{
		if (tmu == GR_TMU1 && nbTextureUnits <= 2) return;
		glActiveTextureARB(GL_TEXTURE0_ARB);

		if (info->aspectRatioLog2 < 0)
		{
			tex0_height = 256;
			tex0_width = tex0_height >> -info->aspectRatioLog2;
		}
		else
		{
			tex0_width = 256;
			tex0_height = tex0_width >> info->aspectRatioLog2;
		}

		glBindTexture(GL_TEXTURE_2D, startAddress+1);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, min_filter0);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mag_filter0);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrap_s0);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrap_t0);
		if(!glsl_support)
		{
			if (need_lambda[0])
				glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, lambda_color[0]);
			else
				glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, texture_env_color);
			updateCombiner(0);
			updateCombinera(0);
		}
	}
	else
	{
		glActiveTextureARB(GL_TEXTURE1_ARB);

		if (info->aspectRatioLog2 < 0)
		{
			tex1_height = 256;
			tex1_width = tex1_height >> -info->aspectRatioLog2;
		}
		else
		{
			tex1_width = 256;
			tex1_height = tex1_width >> info->aspectRatioLog2;
		}

		glBindTexture(GL_TEXTURE_2D, startAddress+1);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, min_filter1);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mag_filter1);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrap_s1);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrap_t1);
		if(!glsl_support)
		{
			if (need_lambda[1])
				glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, lambda_color[1]);
			else
				glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, texture_env_color);
			updateCombiner(1);
			updateCombinera(1);
		}
	}
	if(!CheckTextureBufferFormat(tmu, startAddress+1, info))
	{
		if(tmu == 0 && blackandwhite1 == 1)
		{
			blackandwhite1 = 0;
			need_to_compile = 1;
		}
		if(tmu == 1 && blackandwhite0 == 1)
		{
			blackandwhite0 = 0;
			need_to_compile = 1;
		}
	}
}

FX_ENTRY void FX_CALL 
grTexDetailControl(
                   GrChipID_t tmu,
                   int lod_bias,
                   FxU8 detail_scale,
                   float detail_max
                   )
{
	LOG("grTexDetailControl(%d,%d,%d,%d)\r\n", tmu, lod_bias, detail_scale, detail_max);
	if (lod_bias != 31 && detail_scale != 7)
	{
		if (!lod_bias && !detail_scale && !detail_max) return;
		else
			display_warning("grTexDetailControl : %d, %d, %f", lod_bias, detail_scale, detail_max);
	}
	lambda = detail_max;
	if(lambda > 1.0f)
	{
		lambda = 1.0f - (255.0f - lambda);
	}
	if(lambda > 1.0f) display_warning("lambda:%f", lambda);
	
	if(!glsl_support)
	{
		if (tmu == GR_TMU1 || nbTextureUnits <= 2)
		{
			if (tmu == GR_TMU1 && nbTextureUnits <= 2) return;
			if (need_lambda[0])
			{
				int i;
				glActiveTextureARB(GL_TEXTURE0_ARB);
				for (i=0; i<3; i++) lambda_color[0][i] = texture_env_color[i];
				lambda_color[0][3] = lambda;
				glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, lambda_color[0]);
			}
		}
		else
		{
			if (need_lambda[1])
			{
				int i;
				glActiveTextureARB(GL_TEXTURE1_ARB);
				for (i=0; i<3; i++) lambda_color[1][i] = texture_env_color[i];
				lambda_color[1][3] = lambda;
				glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, lambda_color[1]);
			}
		}
	}
	else
		set_lambda();
}

FX_ENTRY void FX_CALL 
grTexLodBiasValue(GrChipID_t tmu, float bias )
{
	LOG("grTexLodBiasValue(%d,%f)\r\n", tmu, bias);
	if (bias != 0 && bias != 1.0f)
		display_warning("grTexLodBiasValue : %f", bias);
}

FX_ENTRY void FX_CALL 
grTexFilterMode(
                GrChipID_t tmu,
                GrTextureFilterMode_t minfilter_mode,
                GrTextureFilterMode_t magfilter_mode
                )
{
	LOG("grTexFilterMode(%d,%d,%d)\r\n", tmu, minfilter_mode, magfilter_mode);
	if (tmu == GR_TMU1 || nbTextureUnits <= 2)
	{
		if (tmu == GR_TMU1 && nbTextureUnits <= 2) return;
		if (minfilter_mode == GR_TEXTUREFILTER_POINT_SAMPLED) min_filter0 = GL_NEAREST;
		else min_filter0 = GL_LINEAR;

		if (magfilter_mode == GR_TEXTUREFILTER_POINT_SAMPLED) mag_filter0 = GL_NEAREST;
		else mag_filter0 = GL_LINEAR;

		glActiveTextureARB(GL_TEXTURE0_ARB);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, min_filter0);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mag_filter0);
	}
	else
	{
		if (minfilter_mode == GR_TEXTUREFILTER_POINT_SAMPLED) min_filter1 = GL_NEAREST;
		else min_filter1 = GL_LINEAR;

		if (magfilter_mode == GR_TEXTUREFILTER_POINT_SAMPLED) mag_filter1 = GL_NEAREST;
		else mag_filter1 = GL_LINEAR;

		glActiveTextureARB(GL_TEXTURE1_ARB);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, min_filter1);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mag_filter1);
	}
}

FX_ENTRY void FX_CALL 
grTexClampMode(
               GrChipID_t tmu,
               GrTextureClampMode_t s_clampmode,
               GrTextureClampMode_t t_clampmode
               )
{
	LOG("grTexClampMode(%d, %d, %d)\r\n", tmu, s_clampmode, t_clampmode);
	if (tmu == GR_TMU1 || nbTextureUnits <= 2)
	{
		if (tmu == GR_TMU1 && nbTextureUnits <= 2) return;
		switch(s_clampmode)
		{
		case GR_TEXTURECLAMP_WRAP:
			wrap_s0 = GL_REPEAT;
			break;
		case GR_TEXTURECLAMP_CLAMP:
			wrap_s0 = GL_CLAMP_TO_EDGE;
			break;
		case GR_TEXTURECLAMP_MIRROR_EXT:
			wrap_s0 = GL_MIRRORED_REPEAT_ARB;
			break;
		default:
			display_warning("grTexClampMode : unknown s_clampmode : %x", s_clampmode);
		}
		switch(t_clampmode)
		{
		case GR_TEXTURECLAMP_WRAP:
			wrap_t0 = GL_REPEAT;
			break;
		case GR_TEXTURECLAMP_CLAMP:
			wrap_t0 = GL_CLAMP_TO_EDGE;
			break;
		case GR_TEXTURECLAMP_MIRROR_EXT:
			wrap_t0 = GL_MIRRORED_REPEAT_ARB;
			break;
		default:
			display_warning("grTexClampMode : unknown t_clampmode : %x", t_clampmode);
		}
		glActiveTextureARB(GL_TEXTURE0_ARB);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrap_s0);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrap_t0);
	}
	else
	{
		switch(s_clampmode)
		{
		case GR_TEXTURECLAMP_WRAP:
			wrap_s1 = GL_REPEAT;
			break;
		case GR_TEXTURECLAMP_CLAMP:
			wrap_s1 = GL_CLAMP_TO_EDGE;
			break;
		case GR_TEXTURECLAMP_MIRROR_EXT:
			wrap_s1 = GL_MIRRORED_REPEAT_ARB;
			break;
		default:
			display_warning("grTexClampMode : unknown s_clampmode : %x", s_clampmode);
		}
		switch(t_clampmode)
		{
		case GR_TEXTURECLAMP_WRAP:
			wrap_t1 = GL_REPEAT;
			break;
		case GR_TEXTURECLAMP_CLAMP:
			wrap_t1 = GL_CLAMP_TO_EDGE;
			break;
		case GR_TEXTURECLAMP_MIRROR_EXT:
			wrap_t1 = GL_MIRRORED_REPEAT_ARB;
			break;
		default:
			display_warning("grTexClampMode : unknown t_clampmode : %x", t_clampmode);
		}
		glActiveTextureARB(GL_TEXTURE1_ARB);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrap_s1);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrap_t1);
	}
}
